
import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:Trailing')

import os
from struct import unpack
from .. import png
from ..gif import gif as gif_module

def run(image):
    """Extracts trailing data from the image.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR/trailing_data.bin
    """

    global output_file
    output_file = os.path.join(image.veritas.results_directory, "trailing_data.bin")

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.trailing:
        logger.debug('Nothing to do.')
        return

    try:
        if image.file.format == "JPEG":
            jpeg(image)
        elif image.file.format == "TIFF":
            tiff(image)
        elif image.file.format == "PNG":
            png(image)
        elif image.file.format == "BMP":
            bmp(image)
        elif image.file.format == "GIF":
            gif(image)
        else:
            print("Image Trailing: No support yet for format {0}".format(image.file.format))
            return
    except Exception as e:
        print("Image Trailing: Something went wrong... please submit a bug report. Error: {}".format(e))


def gif(image):
        
        # Load up the gif
        g = gif_module(fileName=image.veritas.file_name)
        
        # Parse it
        g.parse()
        
        # Check for excess info
        if len(g.gif) > 0:
                print("Discovered trailing data: {0}".format(g.gif))
                with open(output_file, "wb") as outFile:
                        outFile.write(g.gif)


def png(image):
        pngFile = open(image.veritas.file_name,"rb").read()
        
        # TODO: This isn't 100% accurate. Rework to follow tags correctly.
        pngFile = pngFile.split(b"IEND")[1][4:]
        
        if len(pngFile) != 0:
                print("Discovered Trailing Data:\n{0}".format(pngFile))
                with open(output_file,"wb") as outFile:
                        outFile.write(pngFile)


def bmp(image):
        # This one is pretty easy
        myBMP = open(image.veritas.file_name,"rb").read()
        
        size = unpack("<I",myBMP[2:6])[0]
        
        if len(myBMP[size:]) != 0:
                print("Discovered Trailing Data:\n{0}".format(myBMP[size:]))
                with open(output_file,"wb") as outFile:
                        outFile.write(myBMP[size:])


def tiff(image):
        
        # Read in the file
        with open(image.veritas.file_name,"rb") as myFile:
                steg = myFile.read()
        
        # Parse magic
        if steg[0:4] == b"II*\x00":
                fmt = "<I"
                fmt_s = "<h"
        elif steg[0:4] == b"MM*\x00":
                fmt = ">I"
                fmt_s = ">h"
        else:
                print("Trailing: Error Invalid tiff magic numbers")
                return
        
        # Read header
        ifd = unpack(fmt,steg[4:8])[0]
        
        # Read the number of tags
        nEntries = unpack(fmt_s,steg[ifd:ifd+2])[0]
        
        # We want to find the maximum address
        # Right now our max address is the end of the IFD block
        maxAddr = ifd + (nEntries*0xc) + 2
        
        # Loop through the tags
        for i in range(nEntries):
                # Figure out our current file location
                curAddr = ifd + 2 + (i * 0xc)

                tag = unpack(fmt_s,steg[curAddr:curAddr+2])[0]
                tagType = unpack(fmt_s,steg[curAddr+2:curAddr+4])[0]
                count = unpack(fmt,steg[curAddr+4:curAddr+8])[0]
                
                # print("Tag: {0}\nType: {1}".format(tag,tagType))

                # Tag types of ASCII and Unknown both have offsets associated
                if tagType == 2 or tagType == 7:
                        offset = unpack(fmt,steg[curAddr+8:curAddr+0xc])[0]
                        # See if we have a new winner
                        if (offset + count) > maxAddr:
                                maxAddr = offset + count
                                #print("New max offset")
                        # print("Found new offset: {0}".format(hex(offset)))

        # See if we have data hiding at the end
        if len(steg) > maxAddr:
                print("Trailing Data Discovered... Saving")
                print(steg[maxAddr:])
                with open(output_file,"wb") as outFile:
                        outFile.write(steg[maxAddr:])


def jpeg(image):

        # Official specs here: http://www.w3.org/Graphics/JPEG/itu-t81.pdf      

        # Index for marching through the file   
        i = 0
        
        # These markers don't have a length attribute
        nonLenMarkers = [ b'\xff\xd8', b'\xff\x01', b'\xffd0', b'\xffd1', b'\xffd2', b'\xffd3', b'\xffd4', b'\xffd5', b'\xffd6', b'\xffd7' ]

        # Open up the file
        with open(image.veritas.file_name,"rb") as myFile:
                steg = myFile.read()
        
        while True:
                # Grab the current header
                hdr = steg[i:i+2]
                
                # TODO: Add py logging here
                #print("Found Header: {0}".format(hdr))
                
                # if Start of Image, Temporary Private, Restart, things that don't have an associated length field
                if hdr in nonLenMarkers:
                        # Just move to the next marker
                        i = i + 2
                        continue
                
                # If we've found our way to the end of the jpeg
                if hdr == b'\xff\xd9':
                        #print("Made it to the end!")
                        # Increment 2 so we can check the length
                        i += 2
                        break
                
                # Unpack the length field
                ln = unpack(">H",steg[i+2:i+4])[0]
                
                # print("Found Length: {0}".format(ln))
                
                # Update the index with the known length
                i = i+ln+2
                
                # When we hit scan data, we scan to the end of the format
                if hdr == b'\xff\xda':
                        #print("Start of Scan data")
                        # Find the end marker
                        i += steg[i:].index(b'\xff\xd9')
        
        # Check for trailers
        if i != len(steg):
                print("Trailing Data Discovered... Saving")
                print(steg[i:])
                # Save it off for reference
                with open(output_file, "wb") as outFile:
                        outFile.write(steg[i:])




import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:Meta')

import os
from struct import unpack
import exifread
from .. import png

def run(image):
    """Extracts meta data from the image.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR/metadata
    """

    # This is pretty much replaced entirely by exiftool at the moment..
    return

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.meta:
        logger.debug('Nothing to do.')
        return

    if image.file.format == "PNG":
        PNGMeta(image)

    else:
        logger.debug("No metadata parsing support for {0}".format(image.file.format))

def parsePNGChunk(t,c):
	"""
	Input:
		t -- png chunk type
		c -- png chunk of type bytes
	Action:
		Parse the chunk into something human readable
	Returns:
		string containing relevant information
	"""
	out = ""
	
	if t == "IHDR":
		colortype = {
			0: "Gray scale",
			2: "RGB",
			3: "Color and Palette",
			4: "Gray scale + Alpha",
			6: "Color and Alpha Channel"
		}
		compression = {
			0: "Deflate"
		}
		filtering = {
			0: "Adaptive Filtering"
		}
		interlace = {
			0: "No Interlace",
			1: "Adam7 Interlace"
		}
		u = unpack(">IIbbbbb",c)
		out += "{0:<25}: {1}x{2}\n".format("Size",u[0],u[1])
		out += "{1:<25}: {0}\n".format(u[2],"Bit Depth")
		out += "{1:<25}: {0}\n".format(colortype[u[3]],"Color Type")
		out += "{1:<25}: {0}\n".format(compression[u[4]],"Compression Used")
		out += "{1:<25}: {0}\n".format(filtering[u[5]],"Filter Method")
		out += "{1:<25}: {0}".format(interlace[u[6]],"Interlace Method")
		return out
	if t == "PLTE":
		out += "{1:<25}: {0}".format(int(len(c)/3),"Palettes")
		return out
	if t == "IDAT":
		# Not caring about IDAT for the moment
		return ""
	if t == "IEND":
		return ""
	if t == "tEXt":
		c2 = c.split(b"\x00")
		print("{0:<25}: {1}".format(c2[0].decode('iso-8859-1'),c2[1].decode('iso-8859-1')))
	# Not sure how to parse these yet...
	if t == "iTXt":
		print(c)
		return ""
	if t == "zTXt":
		print(c)
		return ""
	return ""

def JPEGMeta(image):
    """Deprecated -- This is effectively being replaced by the MultiHandler.Exif."""
    meta = ""

    # Open up the file for reading
    with open(image.veritas.file_name,"rb") as jpegFile:
            tags = exifread.process_file(jpegFile)
            
            for tag in tags:
                    meta += "{0}:\t{1}\n".format(tag,tags[tag])
    
    # Show it to the user	
    print("Exif Data\n=========\n{0}\n".format(meta))
    
    # Save it off
    with open(os.path.join(image.veritas.results_directory,"metadata"),"w") as out:
            out.write(meta)
	

def PNGMeta(image):
	
    r = png.Reader(filename=image.veritas.file_name)
    
    for chunk in r.chunks():
            tmp = parsePNGChunk(chunk[0],chunk[1])
            if tmp != "":
                    print(tmp)





import logging
logger = logging.getLogger('StegoVeritas:Modules:Multi:Analysis:XMP')

import os
import random
import hashlib
import libxmp.utils
from prettytable import PrettyTable

from ....helpers import slugify

def run(multi):
    """Checks for any XMP data in the file.

    Args:
        image: MultiHandler class instance

    Returns:
        None

    Saves the result to RESULTSDIR/xmp/
    """

    args = multi.veritas.args

    # Nothing to do
    if not args.auto and not args.xmp:
        logger.debug('Nothing to do.')
        return

    xmp = libxmp.utils.file_to_dict(multi.veritas.file_name)

    # No XMP data here
    if xmp == {}:
        return

    table = PrettyTable(['key','value'])
    xmp_values = []

    for definition, values in xmp.items():
        for value in values:
            table.add_row([repr(value[0]), repr(value[1])])
            xmp_values.append((value[0], value[1]))

    print("XMPP\n====")
    print(table)

    # Save it out
    save_dir = os.path.join(multi.veritas.results_directory, 'xmp')

    os.makedirs(save_dir, exist_ok=True)

    for key,value in xmp_values:
        outfile = os.path.join(save_dir, slugify(key))

        # Slugify means that we might have collisions
        if os.path.exists(outfile):
            logger.warn('XMP outpath already exists, modifying.')
            outfile += '_' + hashlib.md5(str(random.random()).encode()).hexdigest()

        with open(outfile, 'wb') as f:
            f.write(value.encode())
    





import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:ExtractFrames')

import os
from PIL import Image

import png
from copy import copy
from struct import pack, unpack

from apng import APNG

def run(image):
    """Extracts individual frames from an animated gif.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR
    """

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.extract_frames:
        logger.debug('Nothing to do.')
        return

    if image.file.format == 'PNG':
        extract_png(image)

    elif image.file.format == 'GIF':
        extract_gif(image)


def extract_png(image):

    # TODO: Maybe just use APNG: https://pypi.org/project/apng/

    img = APNG.open(image.veritas.file_name)

    if len(img.frames) == 1:
        logger.debug('Only one frame detected.')
        return

    print("Extracting {} PNG frames.".format(len(img.frames)))
    for i, (png, control) in enumerate(img.frames):
        outfile = os.path.join(image.veritas.results_directory, '{}_frame_{}.png'.format(os.path.basename(image.veritas.file_name), i))
        png.save(outfile)

    return

    ### Below is some original work before finding that library
    
    # PIL apparently can't handle this. Neither can ffmpeg. ImageMagick can be fooled.
    # Time to write my own animated png extractor

    img = png.Reader(image.veritas.file_name)
    chunks = list(img.chunks())

    num_frames = len(list(x for x in chunks if x[0] in [b'fcTL']))
    
    """
    actl = next(x for x in chunks if x[0] == b'acTL')

    # How many frames does the image claim to have
    given_num_frames, _ = unpack('>II', actl[1])

    if given_num_frames != num_frames:
        print('Image only claims {} frames, but I discovered {}'.format(given_num_frames, num_frames))
        print('Patching and attempting to extract the frames')

        new_actl = (b'acTL', pack('>II',num_frames,0))

        new_chunks = []
        frame_number = 0
        for t,v in chunks:
            if t == b'acTL':
                new_chunks.append(new_actl)

            elif t == b'fcTL':
                new_chunks.append((t, pack('>I', frame_number) + v[4:]))
                frame_number += 1

            else:
                new_chunks.append((t,v))

        with open('blerg.png','wb') as f:
            png.write_chunks(f, new_chunks)

    return
    """

    #######

    if num_frames == 1:
        logger.debug('Only one frame discovered.')
        return

    print('Discovered multuple PNG frames. Attempting to extract them...')

    header_chunk = next(x for x in chunks if x[0] in [b'IHDR'])
    end_chunk = (b'IEND', b'')
    actl_chunk = (b'acTL', b'\x00\x00\x00\x01\x00\x00\x00\x00') # looping, 1 frame

    new_image = []
    num_frames = 0

    # Type and value for the chunks
    for t, val in chunks:
        print(t)
        
        # We're done with this image
        if t in [b'fcTL', b'IEND']:

            # We have something to save
            if new_image != []:

                # Put propper chunks in place
                new_image.insert(0, copy(actl_chunk))
                new_header_chunk = list(copy(header_chunk))
                #new_header_chunk[1] = pack('>II', width, height) + new_header_chunk[1][8:]
                #new_header_chunk[1] = width_height + new_header_chunk[1][8:]
                new_header_chunk = tuple(new_header_chunk)
                new_image.insert(0, new_header_chunk)
                #new_image.insert(0, copy(header_chunk))
                new_image.append(copy(end_chunk))
                print(new_image)

                outfile = os.path.join(image.veritas.results_directory, '{}_frame_{}.png'.format(os.path.basename(image.veritas.file_name), num_frames))
                print(outfile)

                with open(outfile, 'wb') as f:
                    png.write_chunks(f, new_image)

                #new_image = [(t, val)]
                new_image = []
                num_frames += 1

            if t == b'fcTL':
                # Update our frame size
                #width, height = unpack('>II', val[4:12])
                #print('width and height: ' + str(width) + ' ' + str(height))
                #width_height = val[4:12]
                #new_image.append((t, b'\x00\x00\x00\x00' + val[4:]))
                new_image.append((t, val))

            # Nothing to save, just add our header in
            #else:
            #    new_image.append((t, val))

        elif t == b'IDAT':
            new_image.append((t, val))

        elif t == b'fdAT':
            # fdAT is simply IDAT with 4 bytes at the beginning
            new_image.append((b'IDAT', val[4:]))


        """
        # Chunk data between image designators
        elif t in [b'IDAT', b'fdAT']:
            new_image.append((t, val))
        """



    

def extract_gif(image):

    if not hasattr(image.file, 'is_animated') or image.file.is_animated is False:
        logger.debug('This file is not an animated image.')
        return

    frame_number = 0

    while True:
        outfile = os.path.join(image.veritas.results_directory, '{}_frame_{}.gif'.format(os.path.basename(image.veritas.file_name), frame_number))
        image.file.save(outfile)
        
        frame_number += 1

        try:
            image.file.seek(frame_number)
        except EOFError:
            break

    # Be kind, please rewind.
    image.file.seek(0)



#!/usr/bin/env python3

import logging

# Helper tool to encode something LSB style into an image.

from PIL import Image
import argparse


def parse_args():
    parser = argparse.ArgumentParser(description='Simple tool to add LSB stego into an image.')
    parser.add_argument('-file', type=str, help='If specified, encode this file into the image instead of text.')
    parser.add_argument('-bands', type=str, help='Color order. I.e.: RGB, BGR, GBR, etc', default='RGB')
    parser.add_argument('-output', type=str, default=None, help='Specify the output file name (i.e.: out.png)')
    parser.add_argument('input_file', metavar='file', help='File to add LSB stego into.')
    parser.add_argument('text', nargs='*', help='Text to encode into the file.')
    return parser.parse_args()

def main():
    global img

    args = parse_args()

    # Hide this file inside the input_file
    if args.file:
        with open(args.file, 'rb') as f:
            text = f.read()
    else:
        text = ' '.join(args.text).encode()

    input_file = args.input_file

    img = Image.open(input_file)

    bits = bytes_to_bits(text)

    # TODO: Implement other raster scan orders
    for y in range(img.size[1]):
        for x in range(img.size[0]):
            for band in list(args.bands):
                try:
                    bit_to_encode = next(bits)
                except:
                    break

                encode_pixel(x, y, band, bit_to_encode, bit=0)
        else:
            continue
        break

    # Write it out
    if args.output is not None:
        if args.output.split('.')[-1].lower() not in ['png', 'bmp']:
            logging.warn('Outputting to lossy formats might screw up LSB! Should probably stick to PNG or BMP formats.')

        img.save(args.output)
    else:
        img.save(input_file + '_lsb.png')


def encode_pixel(x, y, band, value, bit=0):
    """Encodes the value (0 or 1) into pixel x,y band on bit 0 (least significant) to 7 (most significant)."""
    assert type(x) is int
    assert type(y) is int
    assert type(band) is str
    assert type(value) is int
    assert value in [0,1]
    assert type(bit) is int
    assert bit in range(8)

    orig_value = get_pixel(x, y, band)
    bitmask = 2**bit

    if value == 0:
        # If the bit is set, unset it
        if orig_value & bitmask != 0:
            new_value = orig_value - bitmask

        # The bit already happened to be correct
        else:
            new_value = orig_value

    else:
        # Or should always bring the value up to 1
        new_value = orig_value | bitmask

    update_pixel(x, y, band, new_value)
    assert get_pixel(x,y,band) == new_value


def get_pixel(x, y, band):
    """Returns the value of band for pixel x,y."""
    assert type(x) is int
    assert type(y) is int
    assert type(band) is str

    band = band.upper()
    band_index = img.getbands().index(band)

    pixels = img.load()
    return pixels[x, y][band_index]


def update_pixel(x, y, band, value):
    """Changes the pixel at x,y band to the given value."""
    assert type(x) is int
    assert type(y) is int
    assert type(band) is str
    assert type(value) is int

    band = band.upper()
    band_index = img.getbands().index(band)

    pixels = img.load()
    pixel = list(pixels[x, y])
    pixel[band_index] = value
    pixels[x, y] = tuple(pixel)

def bytes_to_bits(b, reverse_endian=False):
    if not reverse_endian:
        return map(int, ''.join([bin(i).lstrip('0b').rjust(8,'0') for i in b]))
    else:
        return map(int, ''.join([bin(i).lstrip('0b').rjust(8,'0')[::-1] for i in b]))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from . import Colorer

import logging
logging.basicConfig(level=logging.WARN)

logger = logging.getLogger('StegoVeritas')

import binascii
import os
import argparse
from .version import version

import magic
import time
import shutil
from copy import copy

from .install_deps import required_packages

import binwalk
import tempfile
import hashlib
import random

from prettytable import PrettyTable

class StegoVeritas(object):

    def __init__(self, args=None):
        """
        Args:
            args (list, optional): Arguments as if you passed them via the command line (i.e.: ['-out','directory','-meta'])
        """
        
        self._preflight()
        self._parse_args(args)
        self.modules = []

        # For things we find inside test_output call
        self._keeper_directory = os.path.join(self.results_directory, 'keepers')
        os.makedirs(self._keeper_directory, exist_ok=True)


    def run(self):
        """Run analysis on the file."""
        for module in modules.iter_modules(self):
            print('Running Module: ' + module.__class__.__name__)
            print(module.description)
            module.run()
            self.modules.append(module)

    def test_output(self, thing):
            """
            Args:
                thing (bytes): Renerally from the dump functions
                    ex: thing = b'\x01\x02\x03'

            Returns:
                Nothing. Move output into keep directory if it's worth-while    

            
            Test if output is worth keeping. If it is, move it into the results directory.
            Initially, this is using the Unix file command on the output and checking for non "Data" returns
            """
            
            assert type(thing) == bytes, 'test_output got unexpected thing type of {}'.format(type(thing))


            # TODO: Test new logic...
            # TODO: Iterate through binary offset to find buried data

            #
            # File magic test
            #

            m = magic.from_buffer(thing,mime=True)

            # Generic Output
            if m != 'application/octet-stream':
                m = magic.from_buffer(thing,mime=False)
                print("Found something worth keeping!\n{0}".format(m))
                # Save it to disk
                # TODO: Minor race condition here if we end up multi-processing
                with open(os.path.join(self._keeper_directory, str(time.time())), "wb") as f:
                    f.write(thing)

            #
            # binwalk test
            #

            # TODO: Update this to in-memory scanning if binwalk updates their stuff: https://github.com/ReFirmLabs/binwalk/issues/389
            with tempfile.TemporaryDirectory() as tmpdirname:  
                tmp_scan_file = os.path.join(tmpdirname, 'scanme')

                # Couldn't find a good 'output directory' option for binwalk. Changing dirs because of this.
                saved_dir = os.getcwd()
                os.chdir(tmpdirname)

                with open(tmp_scan_file, 'wb') as f:
                    f.write(thing)

                table = PrettyTable(['Offset', 'Carved/Extracted', 'Description', 'File Name'])
                table.align = 'l'
                keepers = []

                # Run the scan
                for module in binwalk.scan(tmp_scan_file, signature=True, quiet=True, extract=True):
                    for result in module.results:
                        if result.file.path in module.extractor.output:
                            if result.offset in module.extractor.output[result.file.path].carved:
                                table.add_row([hex(result.offset), 'Carved', result.description, os.path.basename(module.extractor.output[result.file.path].carved[result.offset])])
                                keepers.append(module.extractor.output[result.file.path].carved[result.offset])
                                #print("Carved data from offset 0x%X to %s" % (result.offset, module.extractor.output[result.file.path].carved[result.offset]))
                            if result.offset in module.extractor.output[result.file.path].extracted:
                                #print(result.offset, module.extractor.output[result.file.path].extracted)
                                table.add_row([hex(result.offset), 'Extracted', result.description, os.path.basename(module.extractor.output[result.file.path].extracted[result.offset].files[0])])
                                keepers += module.extractor.output[result.file.path].extracted[result.offset].files
                                #print("Extracted %d files from offset 0x%X to '%s' using '%s'" % (len(module.extractor.output[result.file.path].extracted[result.offset].files), result.offset, module.extractor.output[result.file.path].extracted[result.offset].files[0], module.extractor.output[result.file.path].extracted[result.offset].command))

                # If we found something
                if keepers != []:
                    print(table)
                    
                    for keeper in keepers:
                        keeper_dst = os.path.join(self._keeper_directory, os.path.basename(keeper))

                        if os.path.exists(keeper_dst):
                            logger.warn('Keeper name already exists, modifying.')
                            keeper_dst += '_' + hashlib.md5(str(random.random()).encode()).hexdigest()

                        shutil.move( keeper, keeper_dst )

                os.chdir(saved_dir)
            
            # TODO: Check if strings of output contain a known word, save if so.

    def _preflight(self):
        """Checks for missing requirements."""
        
        missing_packages = []

        for package in required_packages:
            if shutil.which(package) is None:
                missing_packages.append(package)

        try:
            import libxmp.utils
        except:
            missing_packages.append('libexempi3')

        if missing_packages != []:
            logger.error('Missing the following required packages: ' + ', '.join(missing_packages))
            logger.error('Either install them manually or run \'stegoveritas_install_deps\'.')

    def _parse_args(self, args=None):

        parser = argparse.ArgumentParser(description='Yet another Stego tool',
                epilog = 'Have a good example? Wish it did something more? Submit a ticket: https://github.com/bannsec/stegoVeritas')

        # Core Options
        parser.add_argument('-out',metavar='dir',type=str, help='Directory to place output in. Defaults to ./results',default=os.path.abspath('./results'))
        parser.add_argument('-debug', action='store_true', help='Enable debugging logging.')
        parser.add_argument('-password',type=str,default=None,help='When applicable, attempt to use this password to extract data.')
        parser.add_argument('-wordlist',type=str,default=None,help='When applicable, attempt to brute force with this wordlist.')
        parser.add_argument('file_name',metavar='file',type=str, default=False, help='The file to analyze')

        # Image Options
        image = parser.add_argument_group('image options')
        image.add_argument('-meta',action='store_true',help='Check file for metadata information')
        image.add_argument('-imageTransform',action='store_true',help='Perform various image transformations on the input image and save them to the output directory')
        image.add_argument('-bruteLSB',action='store_true',help='Attempt to brute force any LSB related stegonography.')
        image.add_argument('-colorMap',nargs="*",metavar='N',type=int, default=None, help='Analyze a color map. Optional arguments are colormap indexes to save while searching')
        image.add_argument('-colorMapRange',nargs=2,metavar=('Start','End'),type=int, default=None, help='Analyze a color map. Same as colorMap but implies a range of colorMap values to keep')
        image.add_argument('-extractLSB',action='store_true',help='Extract a specific LSB RGB from the image. Use with -red, -green, -blue, and -alpha')
        image.add_argument('-red',nargs='+',metavar='index',type=int)
        image.add_argument('-green',nargs='+',metavar='index',type=int)
        image.add_argument('-blue',nargs='+',metavar='index',type=int)
        image.add_argument('-alpha',nargs='+',metavar='index',type=int)
        image.add_argument('-extract_frames',action='store_true',default=False, help='Split up an animated gif into individual frames.')
        image.add_argument('-trailing',action='store_true',help='Check for trailing data on the given file')
        image.add_argument('-steghide',action='store_true',help='Check for StegHide hidden info.')

        # Multi Options
        multi = parser.add_argument_group('multi options')
        multi.add_argument('-exif',action='store_true',default=False,help='Check this file for exif information.')
        multi.add_argument('-xmp',action='store_true',default=False,help='Check this file for XMP information.')
        multi.add_argument('-carve',action='store_true',default=False,help='Attempt to carve/extract things from this file.')

        self.args = parser.parse_args(args)

        if self.args.debug:
            logging.root.setLevel(logging.DEBUG)

        self.file_name = self.args.file_name
        self.results_directory = self.args.out

        # Should this be considered an 'auto' run?
        # TODO: This is SUPER hacky... Should probably find a better way to determine if this is an auto run or not.
        auto = copy(self.args.__dict__)
        auto.pop('out')
        auto.pop('file_name')
        auto.pop('debug')
        self.args.auto = all(option in [False, None, []] for option in auto.values())
    
    ##############
    # Properties #
    ##############

    @property
    def file_name(self) -> str:
        return self.__file_name

    @file_name.setter
    def file_name(self, file_name: str) -> None:
        
        full_path = os.path.abspath(file_name)
        
        if not os.path.exists(full_path):
            logger.error('Cannot find file "{}"'.format(full_path))
            exit(1)

        logger.info('Analyzing file: ' + full_path)
        self.__file_name = full_path

    @property
    def results_directory(self) -> str:
        return self.__results_directory

    @results_directory.setter
    def results_directory(self, results_directory: str) -> None:
        full_path = os.path.abspath(results_directory)

        if os.path.exists(full_path) and not os.path.isdir(full_path):
            logger.error('Output path exists and is not a directory.')
            exit(1)

        os.makedirs(full_path, exist_ok=True)
        
        logger.info('Results Directory: ' + full_path)
        self.__results_directory = results_directory

    @property
    def modules(self) -> list:
        """list: List of all modules that have been run. NOTE: This will only be populated AFTER 'run' has been called and the modules themselves have been run."""
        return self.__modules

    @modules.setter
    def modules(self, modules):
        self.__modules = modules


def main(args=None):
    veritas = StegoVeritas(args=args)
    veritas.run()

from . import modules

if __name__ == '__main__':
    main()




from . import Colorer
import logging
logger = logging.getLogger(__name__)

import subprocess
import distro

import getpass

# What is required by stegoveritas?
required_packages = ['exiftool', '7z', 'foremost']

def main():

    dist_name = distro.name().lower()
    
    if dist_name in ['ubuntu', 'debian', 'kali', 'debian gnu/linux', 'kali gnu/linux']:
        debian()

    elif dist_name == 'fedora':
        fedora()
        
    elif dist_name in ['archlinux', 'arch', 'arch linux', 'manjaro', 'manjaro linux']:
        archlinux()

    else:
        logger.error('Unhandled distribution to install deps: {}'.format(dist_name))
        logger.error('Please poke me or submit a PR.')
        return

def debian():
    
    packages = ['libimage-exiftool-perl', 'libexempi*', 'p7zip-full', 'foremost', 'steghide']

    subprocess.run(command_start + ['apt-get','update'])
    subprocess.run(command_start + ['apt-get','install','-y'] + packages)

def fedora():
    
    packages = ['perl-Image-ExifTool', 'exempi', 'p7zip-plugins', 'foremost', 'steghide']

    subprocess.run(command_start + ['yum','install','-y'] + packages)

def archlinux():
    
    packages = ['perl-image-exiftool', 'p7zip', 'foremost', 'steghide', 'exempi']

    subprocess.run(command_start + ['pacman','-Syu'])
    subprocess.run(command_start + ['pacman','-S'] + packages)

# Standardize using sudo or not
command_start = ['sudo'] if getpass.getuser() != 'root' else []

if __name__ == '__main__':
    main()



from struct import unpack
import sys

class gif:
	"""
	GIF Parser written completely in Python3
	Owner/Copyright: Michael Bann 2015 (bannsecurity.com)
	Originally designed for StegoVeritas tool (https://github.com/Owlz/stegoVeritas)
	"""
	
	# http://www.w3.org/Graphics/GIF/spec-gif89a.txt
		
	# Known gif types
	validGifTypes = [b'GIF87a',b'GIF89a']
	
	def __init__(self,fileName=None):
		"""
		Init the gif class.
		(optional) fileName -- string path to file (i.e.: "mygif.gif" or "/home/blerg/mygif.gif")
		"""
		
		# self.gif == Loaded gif file (not file handle, this is a bytes)
		self.gif = None
		
		# gifType should either be GIF87a or GIF89a
		self.gifType = None

		# Extention info
		self.extentionFunctionLookup = {
			0xfe: self._parseCommentExtension,
			0xf9: self._parseGraphicControlExtension,
			0xff: self._parseApplicationExtension
		}

		# Comments extension
		self.comments = []
		
		# Load up the gif
		self.load(fileName=fileName)
		
		return
	
	def load(self,fileName=None):
		"""
		Input:
			(optional) fileName -- string path to file (i.e.: "mygif.gif" or "/home/blerg/mygif.gif")
		Action:
			Load up the gif specified. If nothing specified, nothing is loaded.
		Return:
			Nothing
		"""
		
		# If we're given the optional gif name param
		if fileName != None:
			self.gif = open(fileName,"rb").read()
		
		return
	
	def _parseHeader(self):
		"""
		Input: None
		Action:
			Parses the header of the already loaded gif and sets class variables.
			Destructive process. Updates the internal gif buffer.
		Returns: None
		"""
		
		t = self.gif[0:6]
		
		# Check if this file has the right header
		if t not in self.validGifTypes:
			raise Exception("Header Error: {0} not a known GIF header".format(t))
		
		self.gifType = t
		
		# Now that we've parsed it, remove it.
		self.gif = self.gif[6:]
	
	def _parseLogicalScreenDescriptor(self):
		"""
		Input: None
		Action:
			Parses the Logical Screen Descriptor section and sets class variables.
			Destructive process. Updates the internal gif buffer.
		Returns: None
		"""

		# Unpack the info		
		self.width, self.height, packed, self.backgroundColorIndex, self.pixelAspectRatio = unpack("<HHBBB",self.gif[:7])
		
		self.globalColorTable = bool((packed & 2**7) >> 7)
		self.colorResolution = ((packed & 0b01110000) >> 4) + 1
		self.sortFlag = bool(((247 & 0b00001000) >> 3))
		self.sizeGlobalColorTable = 2 ** ((247 & 0b00000111) + 1)
		
		# Fix up the aspect ratio if needed
		if self.pixelAspectRatio != 0:
			self.pixelAspectRatio = (self.pixelAspectRatio + 15) / 64
		
		# Update internal gif
		self.gif = self.gif[7:]
		
	def _parseGlobalColorTable(self):
		"""
		Input: None
		Action:
			Parses the Global Color Table. Due to the nature of gif, this needs to be determined by the header first (globalColorTable flag)
		Returns: None
		"""
		
		# Calculate the size
		size = 3 * self.sizeGlobalColorTable
		
		# Just cut out the global color table portion
		gct = self.gif[:size]
		
		self.globalColorTableRed = []
		self.globalColorTableGreen = []
		self.globalColorTableBlue = []
		
		# Loop through the table and save the values off into the respective arrays
		for x in range(0,size,3):
			self.globalColorTableRed.append(gct[x])
			self.globalColorTableGreen.append(gct[x+1])
			self.globalColorTableBlue.append(gct[x+2])
		
		# Update internal gif
		self.gif = self.gif[size:]

	def _parseExtensionBlock(self):
		"""
		Input: None
		Action:
			Assumes that it is called from parse and that the current portion of the input is at an extention block.
			Determines type of extention block and forwards call along to proper handler
		Returns: None
		"""
		
		# Sanity check
		if self.gif[0] != ord("!"):
			raise Exception("_parseExtentionBlock: Something went wrong. We should be at an extension block and we're not")
			return
		
		# Dynamically function calling to make this easier

		if self.gif[1] in self.extentionFunctionLookup:
			self.extentionFunctionLookup[self.gif[1]]()
		else:
			raise Exception("Extension {0} not implimented yet.".format(self.gif[1]))
			return
		
	def _parseCommentExtension(self):
		"""
		Input: None
		Action:
			Parse a comment extension block.
		Returns: None
		"""
		
		# Sanity check
		if self.gif[:2] != b"!\xfe":
			raise Exception("_parseCommentExtension: We're not at a comment extension block.")
			return
		
		# Update our position
		self.gif = self.gif[2:]
		
		length = self.gif[0]
		self.gif = self.gif[1:]
	
		# Loop through all the comments
		while length != 0:
			comment = self.gif[:length]
			# Save the comment off
			self.comments.append(comment)
			self.gif = self.gif[length:]
			length = self.gif[0]
			self.gif = self.gif[1:]
	
	def _parseGraphicControlExtension(self):
		"""
		Input: None
		Action:
			Parse a graphic control extension block
		Return: None
		"""
		
		# Sanity check
		if self.gif[:2] != b"!\xf9":
			raise Exception("_parseGraphicControlExtension: We're not at a graphic control extension block.")
			return
		
		#print("Not parsing GraphicControlExtention for now")
			
		# Skip this for now
		self.gif = self.gif[8:]
	
	def _parseApplicationExtension(self):
		"""
		Input: None
		Action:
			Parse an Application Extension
		Return: None
		"""
		
		# Sanity check
		if self.gif[:2] != b"!\xff":
			raise Exception("_parseApplicationExtension: We're not at an Application Extension block.")
			return
		
		# In actuality this should be static size of 11
		blockSize = self.gif[2]
		
		# ASCII Identifier
		appID = self.gif[3:11]
		appAuthCode = self.gif[11:14]
		
		# Update gif block
		self.gif = self.gif[14:]

		# I only know of one Application Extension... The Netscape extension for looping animations
		if appID == b"NETSCAPE" and appAuthCode == b"2.0":
			"""
			http://www6.uniovi.es/gifanim/gifabout.htm
			byte   1       : 33 (hex 0x21) GIF Extension code
			byte   2       : 255 (hex 0xFF) Application Extension Label
			byte   3       : 11 (hex (0x0B) Length of Application Block 
			                 (eleven bytes of data to follow)
			bytes  4 to 11 : "NETSCAPE"
			bytes 12 to 14 : "2.0"
			byte  15       : 3 (hex 0x03) Length of Data Sub-Block 
			                 (three bytes of data to follow)
			byte  16       : 1 (hex 0x01)
			bytes 17 to 18 : 0 to 65535, an unsigned integer in 
			                 lo-hi byte format. This indicate the 
			                 number of iterations the loop should 
			                 be executed.
			bytes 19       : 0 (hex 0x00) a Data Sub-block Terminator. 
			"""
			# Mainly concerned with the number of times to loop the gif
			self.looping = unpack("<H",self.gif[2:4])[0]
			# Update gif
			self.gif = self.gif[5:]
			return

		else:
			# We should stop parsing at this point since we don't know how to continue
			raise Exception("_parseApplicationExtension: Unknown Application Extension \"{0}\"".format((appID + appAuthCode).decode('ascii')))

	def _parseImageDescriptor(self):
		"""
		Input: None
		Action:
			Parse the image descriptor section of the gif
		Returns: None
		"""
		
		# Sanity check
		if self.gif[0] != ord(","):
			raise Exception("_parseImageDescriptor: File not at Image Descriptor section")
			return
		
		# Parse out the fields
		leftPos,rightPos,imWidth,imHeight,packed = unpack("<HHHHB",self.gif[1:10])
		
		localColorTable = bool(packed & 0b10000000 >> 7)
		sizeLocalColorTable = 2 ** ((packed & 0b00000111)+1)
		
		# Update gif
		self.gif = self.gif[10:]

		# Need extra parsing if this frame is using a local color table
		if localColorTable:
			# Skip it for now
			self.gif = self.gif[sizeLocalColorTable * 3:]
			#print("Skipping local color table parsing for now")
			#raise Exception("_parseImageDescriptor: No support currently for local color tables")
		
		# Not really parsing the data for now. Looping through it
		lzwMinCodeSize = self.gif[0]
		length = self.gif[1]
		self.gif = self.gif[length+2:]
		
		# Loop through all the sub blocks
		while length > 0:
			length = self.gif[0]
			self.gif = self.gif[length+1:]
		


	def parse(self):
		"""
		Input: None
		Action:
			Given the already loaded gif, parse it. Setting internal variables.
		Returns: None
		"""
		
		# Header first
		self._parseHeader()
		
		# Required section -- Logical Screen Descriptor
		self._parseLogicalScreenDescriptor()
		
		# If we are dealing with a global color table
		if self.globalColorTable:
			self._parseGlobalColorTable()
		
		# Now we need to loop through the rest since it's not linear from here
		while len(self.gif) > 0:
			
			# Are we looking at an extention block?
			if self.gif[0] == ord("!"):
				self._parseExtensionBlock()
			# Image Descriptor
			elif self.gif[0] == ord(","):
				self._parseImageDescriptor()
			# Descriptor telling us we're done parsing
			elif self.gif[0] == ord(";"):
				self.gif = self.gif[1:]
				return
			# Something went wrong
			else:
				raise Exception("Unknown identifier: {0}".format(self.gif[0]))

